from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable


@dataclass
class ProgressMessages:
    complete: str = 'Complete'
    error: str = 'Error'
    starting: str = 'Starting...'
    steps: list[str] = field(default_factory=lambda: [
        'Initializing...',
        'Processing...',
        'Analyzing...',
        'Working...',
        'Almost there...',
        'Finalizing...',
    ])


@dataclass
class Task:
    func: Callable
    label: str
    name: str
    parallel: bool = False


@dataclass
class TaskResult:
    error: BaseException | None = None
    value: Any = None

    @property
    def failed(self) -> bool:
        return self.error is not None



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from django_spire.contrib.progress.enums import ProgressStatus

if TYPE_CHECKING:
    from typing import Any


@dataclass
class TaskState:
    message: str = 'Waiting...'
    progress: int = 0
    status: ProgressStatus = ProgressStatus.PENDING

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> TaskState:
        return cls(
            message=data.get('message', 'Waiting...'),
            progress=data.get('progress', 0),
            status=ProgressStatus(data.get('step', ProgressStatus.PENDING)),
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            'message': self.message,
            'progress': self.progress,
            'step': self.status.value,
        }


@dataclass
class TrackerState:
    message: str = 'Initializing...'
    progress: int = 0
    status: ProgressStatus = ProgressStatus.PENDING
    task_order: list[str] = field(default_factory=list)
    tasks: dict[str, TaskState] = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> TrackerState:
        tasks = {
            name: TaskState.from_dict(task_data)
            for name, task_data in data.get('tasks', {}).items()
        }

        return cls(
            message=data.get('message', 'Initializing...'),
            progress=data.get('progress', 0),
            status=ProgressStatus(data.get('step', ProgressStatus.PENDING)),
            task_order=data.get('task_order', []),
            tasks=tasks,
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            'message': self.message,
            'progress': self.progress,
            'step': self.status.value,
            'task_order': self.task_order,
            'tasks': {name: task.to_dict() for name, task in self.tasks.items()},
        }



from __future__ import annotations

import logging
import threading
import time

from typing import TYPE_CHECKING

from django.core.cache import cache

from django_spire.contrib.progress.enums import ProgressStatus
from django_spire.contrib.progress.runner import TaskRunner
from django_spire.contrib.progress.states import TaskState, TrackerState
from django_spire.contrib.progress.task import Task, TaskResult

if TYPE_CHECKING:
    from typing import Any, Callable


log = logging.getLogger(__name__)


class ProgressTracker:
    def __init__(self, key: str, timeout: int = 300) -> None:
        self._key = f'progress_tracker_{key}'
        self._lock = threading.Lock()
        self._tasks: list[Task] = []
        self._timeout = timeout

    def clear(self) -> None:
        cache.delete(self._key)

    def complete(self, message: str = 'Complete!') -> None:
        with self._lock:
            state = self._get_state()
            state.message = message
            state.progress = 100
            state.status = ProgressStatus.COMPLETE
            self._save_state(state)

    def error(self, message: str) -> None:
        with self._lock:
            state = self._get_state()
            state.message = message
            state.progress = 0
            state.status = ProgressStatus.ERROR
            self._save_state(state)

    def execute(self) -> dict[str, Any] | None:
        self._create_state()

        error = False
        results: dict[str, TaskResult] = {}

        try:
            error = self._execute_parallel_tasks(results)

            if not error:
                error = self._execute_sequential_tasks(results)

            if not error:
                self.complete()
                time.sleep(1)
        except BaseException:
            log.exception('Progress tracker failed')
            error = True

        if error:
            self.error('An unexpected error occurred')
            time.sleep(2)

        self.clear()

        if error:
            return None

        return {name: result.value for name, result in results.items()}

    def get(self) -> dict[str, Any] | None:
        return cache.get(self._key)

    def parallel(
        self,
        name: str,
        func: Callable[[], Any],
        label: str | None = None
    ) -> ProgressTracker:
        task = Task(
            func=func,
            label=label or self._generate_label(name),
            name=name,
            parallel=True,
        )

        self._tasks.append(task)

        return self

    def sequential(
        self,
        name: str,
        func: Callable[[dict[str, Any]], Any],
        label: str | None = None
    ) -> ProgressTracker:
        task = Task(
            func=func,
            label=label or self._generate_label(name),
            name=name,
            parallel=False,
        )

        self._tasks.append(task)

        return self

    def start(self) -> None:
        state = TrackerState()
        cache.set(self._key, state.to_dict(), timeout=self._timeout)

    def update(
        self,
        status: ProgressStatus,
        message: str,
        progress: int,
        **kwargs: Any
    ) -> None:
        with self._lock:
            state = self._get_state()
            state.message = message
            state.progress = progress
            state.status = status
            self._save_state(state, **kwargs)

    def _calculate_overall_progress(self, state: TrackerState) -> int:
        if not state.tasks:
            return 0

        total = sum(
            task.progress
            for task in state.tasks.values()
        )

        return int(total / len(state.tasks))

    def _create_state(self) -> None:
        task_names = [task.name for task in self._tasks]

        state = TrackerState(
            task_order=task_names,
            tasks={name: TaskState() for name in task_names}
        )

        cache.set(self._key, state.to_dict(), timeout=self._timeout)

    def _execute_parallel_tasks(self, results: dict[str, TaskResult]) -> bool:
        tasks = [
            task
            for task in self._tasks if task.parallel
        ]

        if not tasks:
            return False

        runners: list[TaskRunner] = []
        threads: list[threading.Thread] = []

        for task in tasks:
            runner = TaskRunner(self, task)
            runners.append(runner)

            thread = threading.Thread(target=runner.run_parallel, args=(results,))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join(timeout=self._timeout)

            if thread.is_alive():
                for runner in runners:
                    runner.stop()

                return True

        return any(
            results.get(task.name, TaskResult()).failed
            for task in tasks
        )

    def _execute_sequential_tasks(self, results: dict[str, TaskResult]) -> bool:
        tasks = [
            task
            for task in self._tasks
            if not task.parallel
        ]

        for task in tasks:
            runner = TaskRunner(self, task)
            runner.run_sequential(results)

            if results.get(task.name, TaskResult()).failed:
                return True

        return False

    def _generate_label(self, name: str) -> str:
        return name.replace('_', ' ').title()

    def _get_state(self) -> TrackerState:
        task_names = [task.name for task in self._tasks]
        data = cache.get(self._key)

        if data is None:
            return TrackerState(
                task_order=task_names,
                tasks={name: TaskState() for name in task_names}
            )

        return TrackerState.from_dict(data)

    def _save_state(self, state: TrackerState, **kwargs: Any) -> None:
        data = state.to_dict()
        data.update(kwargs)
        cache.set(self._key, data, timeout=self._timeout)

    def _update_task(
        self,
        name: str,
        status: ProgressStatus,
        message: str,
        progress: int
    ) -> None:
        with self._lock:
            state = self._get_state()

            state.tasks[name] = TaskState(
                message=message,
                progress=progress,
                status=status,
            )

            state.message = message
            state.progress = self._calculate_overall_progress(state)
            state.status = ProgressStatus.PROCESSING

            self._save_state(state)



from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.contrib.progress.tracker import ProgressTracker

if TYPE_CHECKING:
    from typing import Any

    from django_spire.contrib.progress.enums import ProgressStatus


class ProgressTrackingMixin:
    _tracker: ProgressTracker | None = None

    def get_tracker_key(self) -> str:
        raise NotImplementedError

    @property
    def tracker(self) -> ProgressTracker:
        if self._tracker is None:
            self._tracker = ProgressTracker(self.get_tracker_key())

        return self._tracker

    def progress_error(self, message: str) -> None:
        self.tracker.error(message)

    def update_progress(
        self,
        status: ProgressStatus,
        message: str,
        progress: int,
        **kwargs: Any
    ) -> None:
        self.tracker.update(status, message, progress, **kwargs)



from django_spire.contrib.progress.enums import ProgressStatus
from django_spire.contrib.progress.mixins import ProgressTrackingMixin
from django_spire.contrib.progress.runner import TaskProgressUpdater
from django_spire.contrib.progress.states import TaskState, TrackerState
from django_spire.contrib.progress.task import ProgressMessages, Task, TaskResult
from django_spire.contrib.progress.tracker import ProgressTracker


__all__ = [
    'ProgressMessages',
    'ProgressStatus',
    'ProgressTracker',
    'ProgressTrackingMixin',
    'Task',
    'TaskProgressUpdater',
    'TaskResult',
    'TaskState',
    'TrackerState'
]



from __future__ import annotations

import math
import random
import threading
import time

from typing import TYPE_CHECKING

from django_spire.contrib.progress.enums import ProgressStatus
from django_spire.contrib.progress.task import ProgressMessages, Task, TaskResult

if TYPE_CHECKING:
    from typing import Any, Callable

    from django_spire.contrib.progress.tracker import ProgressTracker


class TaskProgressUpdater:
    def __init__(self, tracker: ProgressTracker, task: Task) -> None:
        self._messages = ProgressMessages()
        self._task = task
        self._tracker = tracker

    def complete(self, message: str | None = None) -> None:
        self._tracker._update_task(
            self._task.name,
            ProgressStatus.COMPLETE,
            self._format(message or self._messages.complete),
            100
        )

    def error(self, message: str) -> None:
        self._tracker._update_task(
            self._task.name,
            ProgressStatus.ERROR,
            self._format(message),
            0
        )

    def start(self) -> None:
        self._tracker._update_task(
            self._task.name,
            ProgressStatus.PROCESSING,
            self._format(self._messages.starting),
            2
        )

    def update(self, message: str, progress: int) -> None:
        self._tracker._update_task(
            self._task.name,
            ProgressStatus.PROCESSING,
            self._format(message),
            progress
        )

    def _format(self, message: str) -> str:
        return f'{self._task.label}: {message}'


class ProgressSimulator:
    def __init__(
        self,
        updater: TaskProgressUpdater,
        max_progress: int = 90,
        update_interval: float = 0.15
    ) -> None:
        self._max_progress = max_progress
        self._messages = ProgressMessages()
        self._update_interval = update_interval
        self._updater = updater

    def run(self, stop_event: threading.Event) -> None:
        start_time = time.time()
        duration = random.uniform(8.0, 15.0)

        while not stop_event.is_set():
            elapsed = time.time() - start_time
            t = min(elapsed / duration, 1.0)

            progress = self._ease_out_expo(t) * self._max_progress
            progress = min(int(progress), self._max_progress)

            jitter = random.uniform(-1.5, 1.5) if progress < self._max_progress - 5 else 0
            progress = max(2, min(int(progress + jitter), self._max_progress))

            message = self._get_message_for_progress(progress)
            self._updater.update(message, progress)

            if progress >= self._max_progress:
                break

            time.sleep(self._update_interval + random.uniform(0, 0.1))

    def _ease_out_expo(self, t: float) -> float:
        if t >= 1.0:
            return 1.0

        return 1.0 - math.pow(2, -10 * t)

    def _get_message_for_progress(self, progress: int) -> str:
        steps = self._messages.steps
        index = min(int(progress / (self._max_progress / len(steps))), len(steps) - 1)

        return steps[index]


class TaskRunner:
    def __init__(self, tracker: ProgressTracker, task: Task) -> None:
        self._stop_event = threading.Event()
        self._task = task
        self._tracker = tracker
        self._updater = TaskProgressUpdater(tracker, task)

    def run_parallel(self, results: dict[str, TaskResult]) -> None:
        self._execute(results, lambda: self._task.func())

    def run_sequential(self, results: dict[str, TaskResult]) -> None:
        unwrapped = {name: result.value for name, result in results.items()}
        self._execute(results, lambda: self._task.func(unwrapped))

    def stop(self) -> None:
        self._stop_event.set()

    def _execute(self, results: dict[str, TaskResult], executor: Callable[[], Any]) -> None:
        simulator = ProgressSimulator(self._updater)
        progress_thread = threading.Thread(target=simulator.run, args=(self._stop_event,))
        progress_thread.start()

        try:
            self._updater.start()
            value = executor()
            results[self._task.name] = TaskResult(value=value)
            self._updater.complete()
        except BaseException as e:
            results[self._task.name] = TaskResult(error=e)
            self._updater.error(str(e))
        finally:
            self._stop_event.set()
            progress_thread.join(timeout=1)
